@@ -4,6 +4,10 @@ use strict;
 use Carp ();
 use String::CRC32 ();
 
+use Gearman::Taskset;
+use Gearman::Util;
+use Storable;
+
 # constructor, given: ($func, $argref, $opts);
 sub new {
     my $class = shift;
@@ -19,14 +23,16 @@ sub new {
 
     my $opts = shift || {};
     for my $k (qw( uniq
-                   on_complete on_fail on_retry on_status
-                   retry_count fail_after_idle high_priority
+                   on_complete on_exception on_fail on_retry on_status
+                   retry_count timeout high_priority background try_timeout
                )) {
         $self->{$k} = delete $opts->{$k};
     }
 
     $self->{retry_count} ||= 0;
 
+    $self->{is_finished} = 0;  # bool: if success or fail has been called yet on this.
+
     if (%{$opts}) {
         Carp::croak("Unknown option(s): " . join(", ", sort keys %$opts));
     }
@@ -36,6 +42,34 @@ sub new {
     return $self;
 }
 
+sub run_hook {
+    my Gearman::Task $self = shift;
+    my $hookname = shift || return;
+
+    my $hook = $self->{hooks}->{$hookname};
+    return unless $hook;
+
+    eval { $hook->(@_) };
+
+    warn "Gearman::Task hook '$hookname' threw error: $@\n" if $@;
+}
+
+sub add_hook {
+    my Gearman::Task $self = shift;
+    my $hookname = shift || return;
+
+    if (@_) {
+        $self->{hooks}->{$hookname} = shift;
+    } else {
+        delete $self->{hooks}->{$hookname};
+    }
+}
+
+sub is_finished {
+    my Gearman::Task $task = $_[0];
+    return $task->{is_finished};
+}
+
 sub taskset {
     my Gearman::Task $task = shift;
 
@@ -78,23 +112,25 @@ sub _hashfunc {
 
 sub pack_submit_packet {
     my Gearman::Task $task = shift;
-    my $is_background = shift;
+    my Gearman::Client $client = shift;
 
-    my $mode = $is_background ?
-        "submit_job_bg" :
-        ($task->{high_priority} ?
-         "submit_job_high" :
-         "submit_job");
+    my $func = $task->{func};
 
-    return Gearman::Util::pack_req_command($mode,
+    if (my $prefix = $client && $client->prefix) {
+        $func = join "\t", $prefix, $task->{func};
+    }
+
+    return Gearman::Util::pack_req_command($task->mode,
                                            join("\0",
-                                                $task->{func} || '',
+                                                $func || '',
                                                 $task->{uniq} || '',
                                                 ${ $task->{argref} } || ''));
 }
 
 sub fail {
     my Gearman::Task $task = shift;
+    my $reason = shift;
+    return if $task->{is_finished};
 
     # try to retry, if we can
     if ($task->{retries_done} < $task->{retry_count}) {
@@ -104,20 +140,50 @@ sub fail {
         return $task->{taskset}->add_task($task);
     }
 
-    return undef unless $task->{on_fail};
-    $task->{on_fail}->();
+    $task->final_fail($reason);
+}
+
+sub final_fail {
+    my Gearman::Task $task = $_[0];
+    my $reason = $_[1];
+
+    return if $task->{is_finished};
+    $task->{is_finished} = $_[1] || 1;
+
+    $task->run_hook('final_fail', $task);
+
+    $task->{on_fail}->($reason) if $task->{on_fail};
+    $task->{on_post_hooks}->()  if $task->{on_post_hooks};
+    $task->wipe;
+
     return undef;
 }
 
+sub exception {
+    my Gearman::Task $task = shift;
+    my $exception_ref = shift;
+    my $exception = Storable::thaw($$exception_ref);
+    $task->{on_exception}->($$exception) if $task->{on_exception};
+    return;
+}
+
 sub complete {
     my Gearman::Task $task = shift;
-    return unless $task->{on_complete};
+    return if $task->{is_finished};
+
     my $result_ref = shift;
-    $task->{on_complete}->($result_ref);
+    $task->{is_finished} = 'complete';
+
+    $task->run_hook('complete', $task);
+
+    $task->{on_complete}->($result_ref) if $task->{on_complete};
+    $task->{on_post_hooks}->()          if $task->{on_post_hooks};
+    $task->wipe;
 }
 
 sub status {
     my Gearman::Task $task = shift;
+    return if $task->{is_finished};
     return unless $task->{on_status};
     my ($nu, $de) = @_;
     $task->{on_status}->($nu, $de);
@@ -131,6 +197,41 @@ sub handle {
     return $task->{handle} = shift;
 }
 
+sub set_on_post_hooks {
+    my Gearman::Task $task = shift;
+    my $code = shift;
+    $task->{on_post_hooks} = $code;
+}
+
+
+sub wipe {
+    my Gearman::Task $task = shift;
+    foreach my $f (qw(on_post_hooks on_complete on_fail on_retry on_status hooks)) {
+        $task->{$f} = undef;
+    }
+}
+
+sub func {
+    my Gearman::Task $task = shift;
+    return $task->{func};
+}
+
+sub timeout {
+    my Gearman::Task $task = shift;
+    return $task->{timeout} unless @_;
+    return $task->{timeout} = shift;
+}
+
+sub mode {
+    my Gearman::Task $task = shift;
+    return $task->{background} ?
+        ($task->{high_priority} ?
+         "submit_job_high_bg" :
+         "submit_job_bg") :
+        ($task->{high_priority} ?
+         "submit_job_high" :
+         "submit_job");
+}
 
 1;
 __END__
@@ -218,11 +319,24 @@ Number of times job will be retried if there are failures.  Defaults to 0.
 Boolean, whether this job should take priority over other jobs already
 enqueued.
 
-=item * fail_after_idle
+=item * timeout
 
-Automatically fail after this many seconds have elapsed.  Defaults to 0,
-which means never.
+Automatically fail, calling your on_fail callback, after this many
+seconds have elapsed without an on_fail or on_complete being
+called. Defaults to 0, which means never.  Bypasses any retry_count
+remaining.
+
+=item * try_timeout
+
+Automatically fail, calling your on_retry callback (or on_fail if out of
+retries), after this many seconds have elapsed. Defaults to 0, which means
+never.
 
 =back
 
+=head2 $task->is_finished
+
+Returns bool: whether or not task is totally done (on_failure or
+on_complete callback has been called)
+
 =cut
